import binascii
import os, re, hashlib
import random
import struct
import sys

import ecdsa
from typing import Union
from .bs58 import b58encode, b58decode, base58_check_encode, base58encodeCheck, base58decode, base58encode
from hdwallet import HDWallet as HD_W
from hdwallet.symbols import BTC, ETH, TRX, LTC, DOGE, DGB, BTG, RVN, QTUM, DASH, ZEC, BCH, AXE
from mnemonic import Mnemonic
from .assest import (
    MAIN_DIGEST_RMD160,
    MAX_PRIVATE_KEY,
    MAIN_PREFIX,
    MAIN_SUFFIX,
    ZERO_BASE_NET,
    VERSION_NETWORK,
    BASE58_ALPHABET,
    FINGERPRINT_RMD160,
    COMPRESSED_PREFIX,
    COMPRESSED_PREFIX2,
    UNCOMPRESSED_PREFIX,
    MAIN_DIGEST,
    XPUB_PREFIX,
    ZERO_BYTES,
    BIP39
)


class Generator:
    def __init__(self):
        super().__init__()

    def checkValid(self, key: int) -> bool:
        if 0 < key < MAX_PRIVATE_KEY:
            return True
        else:
            raise ValueError(f"Secret Scalar Must be greater than 0 and less than {MAX_PRIVATE_KEY}.")

    def generate_private_key(self) -> str:
        randkey = "".join(random.choice("0123456789abcdef") for _ in range(64))
        if self.checkValid(int(randkey, 16)):
            return randkey
        else:
            return self.generate_private_key()

    def generate_xprv(self):
        return "xprv" + binascii.hexlify(os.urandom(32)).decode('utf-8')

    def generate_decimal(self) -> int:
        return random.randint(0, MAX_PRIVATE_KEY)

    def generate_binary(self) -> str:
        return "".join(random.choice("01") for _ in range(256))

    def generate_entropy(self, entropy_bits=256):
        entropy = os.urandom(entropy_bits // 8)
        checksum = hashlib.sha256(entropy).digest()[0]
        entropy_with_checksum = entropy + bytes([checksum])
        return entropy_with_checksum

    def generate_mnemonic(self, size: int) -> str:
        characters = re.findall('[A-Z][a-z]+', BIP39)
        return " ".join(random.choices(characters, k=size)).lower()

    def generate_short_key(self):
        return ('%s%s' % ('S', ''.join(
            [BASE58_ALPHABET[random.randrange(0, len(BASE58_ALPHABET))] for i in range(29)])))


class Convertor:
    def __init__(self):
        super().__init__()
        self.gen = Generator()

    def double_sha256(self, data):
        return hashlib.sha256(hashlib.sha256(data).digest()).digest()

    def mne_to_seed(self, mnemonic, password=""):
        salt = ("mnemonic" + password).encode('utf-8')
        seed = hashlib.pbkdf2_hmac('sha512', mnemonic.encode('utf-8'), salt, 2048)
        return seed[:32]

    def mne_to_bytes(self, mnemonic):
        return self.mne_to_seed(mnemonic)

    def mne_to_hex(self, mnemonic):
        seed = self.mne_to_seed(mnemonic)
        return self.bytes_to_hex(seed)

    def mne_to_wif(self, mnemonic, compress: bool = False):
        seed = self.mne_to_seed(mnemonic)
        return self.bytes_to_wif(seed, compress)

    def mne_to_int(self, mnemonic):
        seed = self.mne_to_seed(mnemonic)
        return self.bytes_to_int(seed)

    def mne_to_xpub(self, mnemonic):
        seed = self.mne_to_seed(mnemonic)
        return self.bytes_to_xpub(seed)

    def mne_to_xprv(self, mnemonic):
        seed = self.mne_to_seed(mnemonic)
        return self.bytes_to_xprv(seed)

    def mne_to_addr(self, mnemonic, compress: bool = False):
        seed = self.mne_to_seed(mnemonic)
        return self.bytes_to_addr(seed, compress)

    def mne_to_binary(self, mnemonic):
        seed = self.mne_to_seed(mnemonic)
        return self.bytes_to_binary(seed)

    def bytes_to_mne(self, seed):
        return Mnemonic().to_mnemonic(seed)

    def bytes_to_seed(self, seed):
        return hashlib.pbkdf2_hmac('sha512', seed, b'mnemonic', 2048)

    def bytes_to_hex(self, seed):
        return binascii.hexlify(self.bytes_to_seed(seed)).decode('utf-8')

    def unHexlify(self, h: str):
        return binascii.unhexlify(h)

    def hex_to_bytes(self, hexed):
        return binascii.unhexlify(hexed)

    def hex_to_mne(self, hexed: str) -> str:
        seed = self.hex_to_bytes(hexed)
        return self.bytes_to_mne(seed)

    def hex_to_wif(self, hexed, compress: bool = False) -> str:
        return self.bytes_to_wif(self.hex_to_bytes(hexed), compress)

    def hex_to_xprv(self, hexed: str) -> str:
        return self.bytes_to_xprv(self.hex_to_bytes(hexed))

    def hex_to_xpub(self, hexed: str) -> str:
        return self.bytes_to_xpub(self.hex_to_bytes(hexed))

    def hex_to_int(self, hexed: str) -> int:
        return int(hexed, 16)

    def hex_to_pub(self, hexed: str, compress: bool = False) -> bytes:
        if compress:
            return self.bytes_to_public(self.hex_to_bytes(hexed), True)
        else:
            return self.bytes_to_public(self.hex_to_bytes(hexed), False)

    def hex_to_addr(self, hexed: str, compress: bool = False) -> str:
        seed = self.hex_to_bytes(hexed)
        if compress:
            return self.bytes_to_addr(seed, True)
        else:
            return self.bytes_to_addr(seed, False)

    def hex_to_binary(self, hexed: str) -> str:
        return self.bytes_to_binary(self.hex_to_bytes(hexed))

    def bytes_to_hex(self, seed):
        privatekey_int = int.from_bytes(hashlib.sha256(seed).digest(), byteorder='big')
        self.gen.checkValid(privatekey_int)
        pvkByte = privatekey_int.to_bytes(32, byteorder='big')
        return pvkByte.hex()

    def bytes_to_int(self, seed) -> int:
        return int.from_bytes(seed, byteorder='big')

    def bytes_to_pub(self, seed_bytes: bytes) -> bytes:
        sk = ecdsa.SigningKey.from_string(seed_bytes[:32], curve=ecdsa.SECP256k1)
        vk = sk.get_verifying_key()
        pub = COMPRESSED_PREFIX2 + vk.to_string()[-32:] if vk.to_string()[-1] % 2 == 0 else b'\x03' + vk.to_string()[
                                                                                                      -32:]
        return pub

    def bytes_to_public(self, seed: bytes, compress: bool = True) -> bytes:
        sk = ecdsa.SigningKey.from_string(seed, curve=ecdsa.SECP256k1)
        vk = sk.get_verifying_key()
        if compress:
            prefix = COMPRESSED_PREFIX2 if vk.pubkey.point.y() % 2 == 0 else COMPRESSED_PREFIX
            return prefix + vk.to_string()[:32]
        else:
            return UNCOMPRESSED_PREFIX + vk.to_string()

    def bytes_to_xpub(self, seed: bytes, chain_code=None) -> str:
        if chain_code is None:
            chain_code = os.urandom(32)  # .hex
        prefix = self.unHexlify(XPUB_PREFIX)
        FINGERPRINT = ZERO_BYTES + ZERO_BYTES
        pub = self.bytes_to_pub(seed)
        xpub = prefix + MAIN_DIGEST + FINGERPRINT + chain_code + pub
        Hash64 = self.double_sha256(xpub)
        xpub += Hash64[:4]
        xpubBase58 = b58encode(xpub)
        return xpubBase58.decode('utf-8')

    def bytes_to_mne(self, byte: bytes):
        seed = byte[:32]
        return Mnemonic("english").to_mnemonic(seed)

    def bytes_to_binary(self, bytes_: bytes) -> str:
        if len(bytes_) != 32:
            raise ValueError("Input bytes should have a length of 32.")

        # Convert each byte to its binary representation and pad with zeros
        return ''.join(format(byte, '08b') for byte in bytes_)

    def bytes_to_wif(self, private_key, compress=True):
        if compress:
            EXTENDED_KEY = MAIN_PREFIX + private_key + MAIN_SUFFIX
        else:
            EXTENDED_KEY = MAIN_PREFIX + private_key

        DOUBLE_SHA256 = self.double_sha256(EXTENDED_KEY)
        CHECKSUM = DOUBLE_SHA256[:4]

        WIF = b58encode(EXTENDED_KEY + CHECKSUM)

        return WIF.decode('utf-8')

    def bytes_to_xprv(self, bytes_code: bytes) -> str:
        chain_code = bytes.fromhex(ZERO_BASE_NET)
        child_number = struct.pack('>L', 0)
        key = MAIN_DIGEST_RMD160 + bytes_code  # 0x00

        xprv_main = VERSION_NETWORK + MAIN_DIGEST_RMD160 + FINGERPRINT_RMD160 + child_number + chain_code + key
        decode_main = base58encodeCheck(b"", xprv_main)
        return decode_main

    def bytes_to_addr(self, seedBytes: bytes, compress: bool = False) -> str:
        if len(seedBytes) != 32:
            seedBytes = seedBytes[:32]
        elif compress:
            pub = self.bytes_to_public(seedBytes, compress=True)
            return self.pub_to_addr(public_key=pub)
        else:
            pub = self.bytes_to_public(seedBytes, compress=False)
            return self.pub_to_addr(public_key=pub)

    # ------------------------------------------------------------
    def pass_to_hex(self, passphrase):
        return hashlib.sha256(passphrase.encode()).hexdigest()

    def pass_to_bytes(self, passphrase: str) -> bytes:
        return bytes.fromhex(self.pass_to_hex(passphrase))

    def pass_to_addr(self, passphrase, compress=False):
        passBytes = self.pass_to_bytes(passphrase)
        sk = ecdsa.SigningKey.from_string(passBytes, curve=ecdsa.SECP256k1)
        vk = sk.verifying_key
        if compress:
            if vk.pubkey.point.y() & 1:
                pub_key = COMPRESSED_PREFIX + vk.to_string()[:32]
            else:
                pub_key = COMPRESSED_PREFIX2 + vk.to_string()[:32]
        else:
            pub_key = UNCOMPRESSED_PREFIX + vk.to_string()
        sha = hashlib.sha256(pub_key).digest()
        ripemd160 = hashlib.new('ripemd160')
        ripemd160.update(sha)

        address = base58_check_encode(ripemd160.digest())
        return "1" + address

    def pass_to_wif(self, passphrase, compress=False):
        passBytes = self.pass_to_bytes(passphrase)
        return self.bytes_to_wif(passBytes, compress)

    def pass_to_xprv(self, passphrase):
        return self.bytes_to_xprv(self.pass_to_bytes(passphrase))

    # ------------------------------------------------------------

    def pub_to_bytes(self, pubkey, compress=True):
        if compress:
            prefix = (COMPRESSED_PREFIX if pubkey.pubkey.point.y() & 1 else COMPRESSED_PREFIX2)
            return prefix + pubkey.pubkey.point.x().to_bytes(32, 'big')
        else:
            point_x = pubkey.pubkey.point.x().to_bytes(32, 'big')
            point_y = pubkey.pubkey.point.y().to_bytes(32, 'big')
            return UNCOMPRESSED_PREFIX + point_x + point_y

    def pub_to_hex(self, pubkey, compress=True):
        return self.pub_to_bytes(pubkey, compress).hex()

    def pub_to_addr(self, public_key: bytes) -> str:
        ripemd160 = hashlib.new('ripemd160')
        ripemd160.update(hashlib.sha256(public_key).digest())
        hashed = MAIN_DIGEST_RMD160 + ripemd160.digest()
        checksum = hashlib.sha256(hashlib.sha256(hashed).digest()).digest()[:4]
        address = hashed + checksum
        return b58encode(address).decode('utf-8')

    # ------------------------------------------------------------

    def wif_to_bytes(self, wif):
        wif_bytes = b58decode(wif)
        isCompress = wif_bytes[-5] == 0x01 if len(wif_bytes) == 38 else False
        return wif_bytes[1:-5] if isCompress else wif_bytes[1:-4]

    def wif_to_binary(self, wif: str) -> str:
        pvkBytes = self.wif_to_bytes(wif)
        return self.bytes_to_binary(pvkBytes)

    def wif_to_addr(self, wif: str, compress: bool = False) -> str:
        pvkBytes = self.wif_to_bytes(wif)
        public_key = self.bytes_to_public(pvkBytes, compress)
        address = self.pub_to_addr(public_key)
        return address

    def wif_to_int(self, wif):
        return self.bytes_to_int(self.wif_to_bytes(wif))

    def wif_to_hex(self, wif):
        return self.wif_to_bytes(wif).hex()

    def wif_to_mne(self, wif):
        return self.bytes_to_mne(self.wif_to_bytes(wif))

    def wif_to_xprv(self, wif):
        return self.bytes_to_xprv(self.wif_to_bytes(wif))

    def wif_to_xpub(self, wif):
        return self.bytes_to_xpub(self.wif_to_bytes(wif))

    def wif_to_pub(self, wif):
        return self.bytes_to_public(self.wif_to_bytes(wif)).hex()

    # ------------------------------------------------------------

    def xprv_to_bytes(self, xprv: str):
        if not xprv.startswith("xprv") or len(xprv) <= 4:
            raise ValueError("Invalid xprv format.")
        xprv58 = xprv[4:]
        xprvBytes = base58decode(xprv58)
        return xprvBytes[:32]

    def xprv_to_addr(self, xprv, compress: bool = False):
        seed = self.xprv_to_bytes(xprv)
        if compress:
            pub = self.bytes_to_public(seed, True)
            return self.pub_to_addr(pub)
        else:
            pub = self.bytes_to_public(seed, False)
            return self.pub_to_addr(pub)

    def xprv_to_pub(self, xprv, compress: bool = False):
        seed = self.xprv_to_bytes(xprv)
        if compress:
            return self.bytes_to_public(seed, True)
        else:
            return self.bytes_to_public(seed, False)

    def xprv_to_wif(self, xprv, compress: bool = False):
        seed = self.xprv_to_bytes(xprv)
        if compress:
            return self.bytes_to_wif(seed, True)
        else:
            return self.bytes_to_wif(seed, False)

    def xprv_to_mne(self, xprv):
        seed = self.xprv_to_bytes(xprv)
        return self.bytes_to_mne(seed)

    # ------------------------------------------------------------

    def binary_to_bytes(self, bin_str: str) -> bytes:
        if len(bin_str) != 256:
            raise ValueError("The binary string must have 256 characters.")
        chunks = [bin_str[i:i + 8] for i in range(0, len(bin_str), 8)]
        return bytes([int(chunk, 2) for chunk in chunks])

    def int_to_bytes(self, int_dec: int) -> bytes:
        return int_dec.to_bytes(32, 'big')

    def int_to_hex(self, int_dec: int) -> str:
        return "%064x" % int_dec

    def int_to_mnemonic(self, int_dec: int) -> str:
        return self.bytes_to_mne(self.int_to_bytes(int_dec))

    def int_to_wif(self, int_dec: int, compress: bool = False) -> str:
        return self.bytes_to_wif(self.int_to_bytes(int_dec), compress)

    def int_to_xprv(self, int_dec: int) -> str:
        return self.bytes_to_xprv(self.int_to_bytes(int_dec))

    def int_to_xpub(self, int_dec: int) -> bytes:
        """
        Convert int decimal to public key (``bytes``).

        :param int_dec:
        :type int_dec: int
        :return:
        :rtype: bytes

        """
        return self.bytes_to_xpub(self.int_to_bytes(int_dec))

    def int_to_addr(self, int_dec: int, compress: bool = False) -> str:
        """
        Convert int decimal to compress & uncompress address (``str``).

        :param int_dec:
        :type int_dec: int
        :param compress:
        :type compress: bool
        :return:
        :rtype: str
        """
        return self.bytes_to_addr(self.int_to_bytes(int_dec), compress)

    def int_to_binary(self, int_dec: int) -> str:
        return self.bytes_to_binary(self.int_to_bytes(int_dec))

    # ------------------------------------------------------------
    def short_to_bytes(self, short_str: str) -> bytes:
        """
        Convert a short string to bytes using SHA-256 hash.

        :param short_str: A short string to be converted.
        :type short_str: str
        :return: A bytes representation of the short string.
        :rtype: bytes
        """
        skey = hashlib.sha256(short_str.encode('utf-8')).hexdigest()
        return binascii.unhexlify(skey)

    def short_to_hex(self, short_str: str) -> str:
        """
        Convert a short key to a hexadecimal string.

        :param short_str: A short key to be converted.
        :type short_str: str
        :return: A hexadecimal string representation of the key.
        :rtype: str
        """
        bkey = self.short_to_bytes(short_str)
        if bkey.startswith(b"\x00"):
            return hashlib.sha256(short_str.encode('utf-8')).hexdigest()
        else:
            return self.bytes_to_hex(bkey)

    def short_to_int(self, short_str: str) -> int:
        """
        Convert a short string to an integer using SHA-256 hash.

        :param short_str: A short string to be converted.
        :type short_str: str
        :return: An integer representation of the short string.
        :rtype: int
        """
        skey = hashlib.sha256(short_str.encode('utf-8')).hexdigest()
        return int(skey, 16)

    def short_to_addr(self, short_str: str, compress: bool = False) -> str:
        hex_key = self.short_to_hex(short_str)
        dec_key = int(hex_key, 16)
        if dec_key <= MAX_PRIVATE_KEY:
            return self.hex_to_addr(hex_key, compress)
        else:
            return self.int_to_addr(dec_key, compress)

    def short_to_wif(self, short_str: str, compress: bool = False) -> str:
        return self.hex_to_wif(self.short_to_hex(short_str), compress)
